@@ -1,5 +1,6 @@
-/* Copyright (C) 2002-2023 Free Software Foundation, Inc.
+/* Copyright (C) 2002, 2003, 2006 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
+   Contributed by Ulrich Drepper <drepper@redhat.com>, 2002.
 
    The GNU C Library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -12,13 +13,15 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the GNU C Library; if not, see
-   <https://www.gnu.org/licenses/>.  */
+   License along with the GNU C Library; if not, write to the Free
+   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   02111-1307 USA.  */
 
+#include <assert.h>
 #include <errno.h>
 #include <limits.h>
+#include <eglibc/dynconf.h>
 #include "pthreadP.h"
-#include <shlib-compat.h>
 
 
 #ifndef NEW_VERNUM
@@ -27,53 +30,47 @@
 
 
 int
-__pthread_attr_setstack (pthread_attr_t *attr, void *stackaddr,
-			 size_t stacksize)
+__pthread_attr_setstack (attr, stackaddr, stacksize)
+     pthread_attr_t *attr;
+     void *stackaddr;
+     size_t stacksize;
 {
   struct pthread_attr *iattr;
 
+  assert (sizeof (*attr) >= sizeof (struct pthread_attr));
   iattr = (struct pthread_attr *) attr;
 
   /* Catch invalid sizes.  */
-  int ret = check_stacksize_attr (stacksize);
-  if (ret)
-    return ret;
+  if (stacksize < EGLIBC_THREAD_STACK_MIN)
+    return EINVAL;
 
 #ifdef EXTRA_PARAM_CHECKS
   EXTRA_PARAM_CHECKS;
 #endif
 
   iattr->stacksize = stacksize;
-#if _STACK_GROWS_DOWN
   iattr->stackaddr = (char *) stackaddr + stacksize;
-#else
-  iattr->stackaddr = (char *) stackaddr;
-#endif
   iattr->flags |= ATTR_FLAG_STACKADDR;
 
   return 0;
 }
-versioned_symbol (libc, __pthread_attr_setstack, pthread_attr_setstack,
-		  GLIBC_2_34);
 
 #if PTHREAD_STACK_MIN == 16384
-# if OTHER_SHLIB_COMPAT (libpthread, GLIBC_2_2, GLIBC_2_34)
-compat_symbol (libpthread, __pthread_attr_setstack, pthread_attr_setstack,
-	       GLIBC_2_2);
-# endif
-#else /* PTHREAD_STACK_MIN != 16384 */
-# if OTHER_SHLIB_COMPAT (libpthread, NEW_VERNUM, GLIBC_2_34)
-compat_symbol (libpthread, __pthread_attr_setstack, pthread_attr_setstack,
-	       NEW_VERNUM);
-# endif
+strong_alias (__pthread_attr_setstack, pthread_attr_setstack)
+#else
+# include <shlib-compat.h>
+versioned_symbol (libpthread, __pthread_attr_setstack, pthread_attr_setstack,
+		  NEW_VERNUM);
+
+# if SHLIB_COMPAT(libpthread, GLIBC_2_2, NEW_VERNUM)
 
-# if OTHER_SHLIB_COMPAT (libpthread, GLIBC_2_2, NEW_VERNUM)
 int
 __old_pthread_attr_setstack (pthread_attr_t *attr, void *stackaddr,
 			     size_t stacksize)
 {
   struct pthread_attr *iattr;
 
+  assert (sizeof (*attr) >= sizeof (struct pthread_attr));
   iattr = (struct pthread_attr *) attr;
 
   /* Catch invalid sizes.  */
@@ -85,11 +82,7 @@ __old_pthread_attr_setstack (pthread_attr_t *attr, void *stackaddr,
 #  endif
 
   iattr->stacksize = stacksize;
-#  if _STACK_GROWS_DOWN
   iattr->stackaddr = (char *) stackaddr + stacksize;
-#  else
-  iattr->stackaddr = (char *) stackaddr;
-#  endif
   iattr->flags |= ATTR_FLAG_STACKADDR;
 
   return 0;
@@ -97,5 +90,6 @@ __old_pthread_attr_setstack (pthread_attr_t *attr, void *stackaddr,
 
 compat_symbol (libpthread, __old_pthread_attr_setstack, pthread_attr_setstack,
 	       GLIBC_2_2);
-# endif /* OTHER_SHLIB_COMPAT */
-#endif /* PTHREAD_STACK_MIN != 16384 */
+# endif
+
+#endif
