@@ -1,4 +1,4 @@
-/* Copyright (C) 2002-2016 Free Software Foundation, Inc.
+/* Copyright (C) 2002, 2003, 2006 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
    Contributed by Ulrich Drepper <drepper@redhat.com>, 2002.
 
@@ -13,12 +13,14 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the GNU C Library; if not, see
-   <http://www.gnu.org/licenses/>.  */
+   License along with the GNU C Library; if not, write to the Free
+   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   02111-1307 USA.  */
 
 #include <assert.h>
 #include <errno.h>
 #include <limits.h>
+#include <eglibc/dynconf.h>
 #include "pthreadP.h"
 
 #ifndef NEW_VERNUM
@@ -27,7 +29,9 @@
 
 
 int
-__pthread_attr_setstacksize (pthread_attr_t *attr, size_t stacksize)
+__pthread_attr_setstacksize (attr, stacksize)
+     pthread_attr_t *attr;
+     size_t stacksize;
 {
   struct pthread_attr *iattr;
 
@@ -35,9 +39,8 @@ __pthread_attr_setstacksize (pthread_attr_t *attr, size_t stacksize)
   iattr = (struct pthread_attr *) attr;
 
   /* Catch invalid sizes.  */
-  int ret = check_stacksize_attr (stacksize);
-  if (ret)
-    return ret;
+  if (stacksize < EGLIBC_THREAD_STACK_MIN)
+    return EINVAL;
 
   iattr->stacksize = stacksize;
 
