@@ -13,17 +13,18 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the GNU C Library; see the file COPYING.LIB.  If
-   not, see <http://www.gnu.org/licenses/>.  */
+   License along with the GNU C Library; see the file COPYING.LIB.  If not,
+   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+   Boston, MA 02111-1307, USA.  */
 
 #include <alloca.h>
 #include <limits.h>
-
+#include <eglibc/dynconf.h>
 
 extern int
 __always_inline
 __libc_use_alloca (size_t size)
 {
-  return (__builtin_expect (size <= PTHREAD_STACK_MIN / 4, 1)
+  return (__builtin_expect (size <= EGLIBC_THREAD_STACK_MIN / 4, 1)
 	  || __builtin_expect (__libc_alloca_cutoff (size), 1));
 }
