@@ -1,4 +1,4 @@
-/* Copyright (C) 2006, 2007, 2008, 2011 Free Software Foundation, Inc.
+/* Copyright (C) 2006, 2007, 2008 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
 
    The GNU C Library is free software; you can redistribute it and/or
@@ -12,8 +12,9 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the GNU C Library; if not, see
-   <http://www.gnu.org/licenses/>.  */
+   License along with the GNU C Library; if not, write to the Free
+   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   02111-1307 USA.  */
 
 /* We define a special synchronization primitive for AIO.  POSIX
    conditional variables would be ideal but the pthread_cond_*wait
@@ -24,6 +25,7 @@
 #include <signal.h>
 #include <pthreadP.h>
 #include <lowlevellock.h>
+#include <eglibc/dynconf.h>
 
 #define DONT_NEED_GAI_MISC_COND	1
 
@@ -96,9 +98,7 @@ __gai_create_helper_thread (pthread_t *threadp, void *(*tf) (void *),
   pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
 
   /* The helper thread needs only very little resources.  */
-  (void) pthread_attr_setstacksize (&attr,
-				    __pthread_get_minstack (&attr)
-				    + 4 * PTHREAD_STACK_MIN);
+  (void) pthread_attr_setstacksize (&attr, 4 * EGLIBC_THREAD_STACK_MIN);
 
   /* Block all signals in the helper thread.  To do this thoroughly we
      temporarily have to block all signals here.  */
