@@ -1,6 +1,7 @@
 /* POSIX.2 wordexp implementation.
-   Copyright (C) 1997-2026 Free Software Foundation, Inc.
+   Copyright (C) 1997-2003, 2005, 2006, 2008 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
+   Contributed by Tim Waugh <tim@cyberelk.demon.co.uk>.
 
    The GNU C Library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -13,9 +14,11 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the GNU C Library; if not, see
-   <https://www.gnu.org/licenses/>.  */
+   License along with the GNU C Library; if not, write to the Free
+   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   02111-1307 USA.  */
 
+#include <alloca.h>
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -24,18 +27,33 @@
 #include <libintl.h>
 #include <paths.h>
 #include <pwd.h>
-#include <stdbool.h>
+#include <signal.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <sys/param.h>
+#include <sys/stat.h>
+#include <sys/time.h>
+#include <sys/types.h>
+#include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
+#ifdef USE_IN_LIBIO
+# include <wchar.h>
+#endif
 #include <wordexp.h>
-#include <spawn.h>
-#include <scratch_buffer.h>
-#include <_itoa.h>
+#include <kernel-features.h>
+
+#include <bits/libc-lock.h>
+#include <stdio-common/_itoa.h>
+
+/* Undefine the following line for the production version.  */
+/* #define NDEBUG 1 */
 #include <assert.h>
 
+/* Get some device information.  */
+#include <device-nrs.h>
+
 /*
  * This is a recursive-descent-style word expansion routine.
  */
@@ -48,16 +66,19 @@ extern char **__libc_argv attribute_hidden;
 static int parse_dollars (char **word, size_t *word_length, size_t *max_length,
 			  const char *words, size_t *offset, int flags,
 			  wordexp_t *pwordexp, const char *ifs,
-			  const char *ifs_white, int quoted);
+			  const char *ifs_white, int quoted)
+     internal_function;
 static int parse_backtick (char **word, size_t *word_length,
 			   size_t *max_length, const char *words,
 			   size_t *offset, int flags, wordexp_t *pwordexp,
-			   const char *ifs, const char *ifs_white);
+			   const char *ifs, const char *ifs_white)
+     internal_function;
 static int parse_dquote (char **word, size_t *word_length, size_t *max_length,
 			 const char *words, size_t *offset, int flags,
 			 wordexp_t *pwordexp, const char *ifs,
-			 const char *ifs_white);
-static int eval_expr (char *expr, long int *result);
+			 const char *ifs_white)
+     internal_function;
+static int eval_expr (char *expr, long int *result) internal_function;
 
 /* The w_*() functions manipulate word lists. */
 
@@ -98,6 +119,7 @@ w_addchar (char *buffer, size_t *actlen, size_t *maxlen, char ch)
 }
 
 static char *
+internal_function
 w_addmem (char *buffer, size_t *actlen, size_t *maxlen, const char *str,
 	  size_t len)
 {
@@ -124,6 +146,7 @@ w_addmem (char *buffer, size_t *actlen, size_t *maxlen, const char *str,
 }
 
 static char *
+internal_function
 w_addstr (char *buffer, size_t *actlen, size_t *maxlen, const char *str)
      /* (lengths exclude trailing zero) */
 {
@@ -138,6 +161,7 @@ w_addstr (char *buffer, size_t *actlen, size_t *maxlen, const char *str)
 }
 
 static int
+internal_function
 w_addword (wordexp_t *pwordexp, char *word)
 {
   /* Add a word to the wordlist */
@@ -178,6 +202,7 @@ no_space:
  */
 
 static int
+internal_function
 parse_backslash (char **word, size_t *word_length, size_t *max_length,
 		 const char *words, size_t *offset)
 {
@@ -206,6 +231,7 @@ parse_backslash (char **word, size_t *word_length, size_t *max_length,
 }
 
 static int
+internal_function
 parse_qtd_backslash (char **word, size_t *word_length, size_t *max_length,
 		     const char *words, size_t *offset)
 {
@@ -248,6 +274,7 @@ parse_qtd_backslash (char **word, size_t *word_length, size_t *max_length,
 }
 
 static int
+internal_function
 parse_tilde (char **word, size_t *word_length, size_t *max_length,
 	     const char *words, size_t *offset, size_t wordc)
 {
@@ -269,8 +296,8 @@ parse_tilde (char **word, size_t *word_length, size_t *max_length,
 
   for (i = 1 + *offset; words[i]; i++)
     {
-      if (words[i] == ':' || words[i] == '/' || words[i] == ' '
-	  || words[i] == '\t' || words[i] == 0 )
+      if (words[i] == ':' || words[i] == '/' || words[i] == ' ' ||
+	  words[i] == '\t' || words[i] == 0 )
 	break;
 
       if (words[i] == '\\')
@@ -283,7 +310,12 @@ parse_tilde (char **word, size_t *word_length, size_t *max_length,
   if (i == 1 + *offset)
     {
       /* Tilde appears on its own */
+      uid_t uid;
+      struct passwd pwd, *tpwd;
+      int buflen = 1000;
       char* home;
+      char* buffer;
+      int result;
 
       /* POSIX.2 says ~ expands to $HOME and if HOME is unset the
 	 results are unspecified.  We do a lookup on the uid if
@@ -298,38 +330,25 @@ parse_tilde (char **word, size_t *word_length, size_t *max_length,
 	}
       else
 	{
-	  struct passwd pwd, *tpwd;
-	  uid_t uid = __getuid ();
-	  int result;
-	  struct scratch_buffer tmpbuf;
-	  scratch_buffer_init (&tmpbuf);
+	  uid = __getuid ();
+	  buffer = __alloca (buflen);
 
-	  while ((result = __getpwuid_r (uid, &pwd,
-					 tmpbuf.data, tmpbuf.length,
-					 &tpwd)) != 0
+	  while ((result = __getpwuid_r (uid, &pwd, buffer, buflen, &tpwd)) != 0
 		 && errno == ERANGE)
-	    if (!scratch_buffer_grow (&tmpbuf))
-	      return WRDE_NOSPACE;
+	    buffer = extend_alloca (buffer, buflen, buflen + 1000);
 
 	  if (result == 0 && tpwd != NULL && pwd.pw_dir != NULL)
 	    {
 	      *word = w_addstr (*word, word_length, max_length, pwd.pw_dir);
 	      if (*word == NULL)
-		{
-		  scratch_buffer_free (&tmpbuf);
-		  return WRDE_NOSPACE;
-		}
+		return WRDE_NOSPACE;
 	    }
 	  else
 	    {
 	      *word = w_addchar (*word, word_length, max_length, '~');
 	      if (*word == NULL)
-		{
-		  scratch_buffer_free (&tmpbuf);
-		  return WRDE_NOSPACE;
-		}
+		return WRDE_NOSPACE;
 	    }
-	  scratch_buffer_free (&tmpbuf);
 	}
     }
   else
@@ -337,15 +356,13 @@ parse_tilde (char **word, size_t *word_length, size_t *max_length,
       /* Look up user name in database to get home directory */
       char *user = strndupa (&words[1 + *offset], i - (1 + *offset));
       struct passwd pwd, *tpwd;
+      int buflen = 1000;
+      char* buffer = __alloca (buflen);
       int result;
-      struct scratch_buffer tmpbuf;
-      scratch_buffer_init (&tmpbuf);
 
-      while ((result = __getpwnam_r (user, &pwd, tmpbuf.data, tmpbuf.length,
-				     &tpwd)) != 0
+      while ((result = __getpwnam_r (user, &pwd, buffer, buflen, &tpwd)) != 0
 	     && errno == ERANGE)
-	if (!scratch_buffer_grow (&tmpbuf))
-	  return WRDE_NOSPACE;
+	buffer = extend_alloca (buffer, buflen, buflen + 1000);
 
       if (result == 0 && tpwd != NULL && pwd.pw_dir)
 	*word = w_addstr (*word, word_length, max_length, pwd.pw_dir);
@@ -357,8 +374,6 @@ parse_tilde (char **word, size_t *word_length, size_t *max_length,
 	    *word = w_addstr (*word, word_length, max_length, user);
 	}
 
-      scratch_buffer_free (&tmpbuf);
-
       *offset = i - 1;
     }
   return *word ? 0 : WRDE_NOSPACE;
@@ -366,6 +381,7 @@ parse_tilde (char **word, size_t *word_length, size_t *max_length,
 
 
 static int
+internal_function
 do_parse_glob (const char *glob_word, char **word, size_t *word_length,
 	       size_t *max_length, wordexp_t *pwordexp, const char *ifs,
 	       const char *ifs_white)
@@ -422,6 +438,7 @@ do_parse_glob (const char *glob_word, char **word, size_t *word_length,
 }
 
 static int
+internal_function
 parse_glob (char **word, size_t *word_length, size_t *max_length,
 	    const char *words, size_t *offset, int flags,
 	    wordexp_t *pwordexp, const char *ifs, const char *ifs_white)
@@ -517,6 +534,7 @@ tidy_up:
 }
 
 static int
+internal_function
 parse_squote (char **word, size_t *word_length, size_t *max_length,
 	      const char *words, size_t *offset)
 {
@@ -538,6 +556,7 @@ parse_squote (char **word, size_t *word_length, size_t *max_length,
 
 /* Functions to evaluate an arithmetic expression */
 static int
+internal_function
 eval_expr_val (char **expr, long int *result)
 {
   char *digit;
@@ -572,6 +591,7 @@ eval_expr_val (char **expr, long int *result)
 }
 
 static int
+internal_function
 eval_expr_multdiv (char **expr, long int *result)
 {
   long int arg;
@@ -599,10 +619,6 @@ eval_expr_multdiv (char **expr, long int *result)
 	  if (eval_expr_val (expr, &arg) != 0)
 	    return WRDE_SYNTAX;
 
-	  /* Division by zero or integer overflow.  */
-	  if (arg == 0 || (arg == -1 && *result == LONG_MIN))
-	    return WRDE_SYNTAX;
-
 	  *result /= arg;
 	}
       else break;
@@ -612,6 +628,7 @@ eval_expr_multdiv (char **expr, long int *result)
 }
 
 static int
+internal_function
 eval_expr (char *expr, long int *result)
 {
   long int arg;
@@ -648,6 +665,7 @@ eval_expr (char *expr, long int *result)
 }
 
 static int
+internal_function
 parse_arith (char **word, size_t *word_length, size_t *max_length,
 	     const char *words, size_t *offset, int flags, int bracket)
 {
@@ -720,7 +738,7 @@ parse_arith (char **word, size_t *word_length, size_t *max_length,
 	      ++(*offset);
 
 	      /* Go - evaluate. */
-	      if (expr && eval_expr (expr, &numresult) != 0)
+	      if (*expr && eval_expr (expr, &numresult) != 0)
 		{
 		  free (expr);
 		  return WRDE_SYNTAX;
@@ -758,7 +776,7 @@ parse_arith (char **word, size_t *word_length, size_t *max_length,
 	      long int numresult = 0;
 
 	      /* Go - evaluate. */
-	      if (expr && eval_expr (expr, &numresult) != 0)
+	      if (*expr && eval_expr (expr, &numresult) != 0)
 		{
 		  free (expr);
 		  return WRDE_SYNTAX;
@@ -783,7 +801,6 @@ parse_arith (char **word, size_t *word_length, size_t *max_length,
 
 	case '(':
 	  ++paren_depth;
-	  [[fallthrough]];
 	default:
 	  expr = w_addchar (expr, &expr_length, &expr_maxlen, words[*offset]);
 	  if (expr == NULL)
@@ -796,81 +813,74 @@ parse_arith (char **word, size_t *word_length, size_t *max_length,
   return WRDE_SYNTAX;
 }
 
-#define DYNARRAY_STRUCT        strlist
-#define DYNARRAY_ELEMENT       char *
-#define DYNARRAY_PREFIX        strlist_
-/* Allocates about 512/1024 (32/64 bit) on stack.  */
-#define DYNARRAY_INITIAL_SIZE  128
-#include <malloc/dynarray-skeleton.c>
-
 /* Function called by child process in exec_comm() */
-static pid_t
-exec_comm_child (char *comm, int *fildes, bool showerr, bool noexec)
+static inline void
+internal_function __attribute__ ((always_inline))
+exec_comm_child (char *comm, int *fildes, int showerr, int noexec)
 {
-  pid_t pid = -1;
+  const char *args[4] = { _PATH_BSHELL, "-c", comm, NULL };
 
-  /* Execute the command, or just check syntax?  */
-  const char *args[] = { _PATH_BSHELL, noexec ? "-nc" : "-c", comm, NULL };
+  /* Execute the command, or just check syntax? */
+  if (noexec)
+    args[1] = "-nc";
 
-  posix_spawn_file_actions_t fa;
-  /* posix_spawn_file_actions_init does not fail.  */
-  __posix_spawn_file_actions_init (&fa);
-
-  /* Redirect output.  For check syntax only (noexec being true), exec_comm
-     explicitly sets fildes[1] to -1, so check its value to avoid a failure
-     in __posix_spawn_file_actions_adddup2.  */
-  if (fildes[1] != -1)
+  /* Redirect output.  */
+  if (__builtin_expect (fildes[1] != STDOUT_FILENO, 1))
     {
-      if (__glibc_likely (fildes[1] != STDOUT_FILENO))
-	{
-	  if (__posix_spawn_file_actions_adddup2 (&fa, fildes[1],
-						  STDOUT_FILENO) != 0
-	      || __posix_spawn_file_actions_addclose (&fa, fildes[1]) != 0)
-	    goto out;
-	}
-      else
-	/* Reset the close-on-exec flag (if necessary).  */
-	if (__posix_spawn_file_actions_adddup2 (&fa, fildes[1], fildes[1])
-	    != 0)
-	  goto out;
+      __dup2 (fildes[1], STDOUT_FILENO);
+      __close (fildes[1]);
+    }
+  else
+    {
+#ifdef O_CLOEXEC
+      /* Reset the close-on-exec flag (if necessary).  */
+# ifndef __ASSUME_PIPE2
+      if (__have_pipe2 > 0)
+# endif
+	__fcntl (fildes[1], F_SETFD, 0);
+#endif
     }
 
   /* Redirect stderr to /dev/null if we have to.  */
-  if (!showerr)
-    if (__posix_spawn_file_actions_addopen (&fa, STDERR_FILENO, _PATH_DEVNULL,
-					    O_WRONLY, 0) != 0)
-      goto out;
-
-  struct strlist newenv;
-  strlist_init (&newenv);
-
-  bool recreate_env = getenv ("IFS") != NULL;
-  if (recreate_env)
+  if (showerr == 0)
     {
-      for (char **ep = __environ; *ep != NULL; ep++)
-	if (strncmp (*ep, "IFS=", strlen ("IFS=")) != 0)
-	  strlist_add (&newenv, *ep);
-      strlist_add (&newenv, NULL);
-      if (strlist_has_failed (&newenv))
-	goto out;
+      struct stat64 st;
+      int fd;
+      __close (STDERR_FILENO);
+      fd = __open (_PATH_DEVNULL, O_WRONLY);
+      if (fd >= 0 && fd != STDERR_FILENO)
+	{
+	  __dup2 (fd, STDERR_FILENO);
+	  __close (fd);
+	}
+      /* Be paranoid.  Check that we actually opened the /dev/null
+         device.  */
+      if (__builtin_expect (__fxstat64 (_STAT_VER, STDERR_FILENO, &st), 0) != 0
+	  || __builtin_expect (S_ISCHR (st.st_mode), 1) == 0
+#if defined DEV_NULL_MAJOR && defined DEV_NULL_MINOR
+	  || st.st_rdev != makedev (DEV_NULL_MAJOR, DEV_NULL_MINOR)
+#endif
+	  )
+	/* It's not the /dev/null device.  Stop right here.  The
+           problem is: how do we stop?  We use _exit() with an
+           hopefully unusual exit code.  */
+	_exit (90);
     }
 
-  /* pid is not set if posix_spawn fails, so it keep the original value
-     of -1.  */
-  __posix_spawn (&pid, _PATH_BSHELL, &fa, NULL, (char *const *) args,
-		 recreate_env ? strlist_begin (&newenv) : __environ);
+  /* Make sure the subshell doesn't field-split on our behalf. */
+  __unsetenv ("IFS");
 
-  strlist_free (&newenv);
+  __close (fildes[0]);
+  __execve (_PATH_BSHELL, (char *const *) args, __environ);
 
-out:
-  __posix_spawn_file_actions_destroy (&fa);
-
-  return pid;
+  /* Bad.  What now?  */
+  abort ();
 }
 
 /* Function to execute a command and retrieve the results */
 /* pwordexp contains NULL if field-splitting is forbidden */
 static int
+internal_function
 exec_comm (char *comm, char **word, size_t *word_length, size_t *max_length,
 	   int flags, wordexp_t *pwordexp, const char *ifs,
 	   const char *ifs_white)
@@ -883,29 +893,55 @@ exec_comm (char *comm, char **word, size_t *word_length, size_t *max_length,
   size_t maxnewlines = 0;
   char buffer[bufsize];
   pid_t pid;
-  bool noexec = false;
+  int noexec = 0;
 
-  /* Do nothing if command substitution should not succeed.  */
-  if (flags & WRDE_NOCMD)
-    return WRDE_CMDSUB;
-
-  /* Don't posix_spawn unless necessary */
+  /* Don't fork() unless necessary */
   if (!comm || !*comm)
     return 0;
 
-  if (__pipe2 (fildes, O_CLOEXEC) < 0)
-    return WRDE_NOSPACE;
+#ifdef O_CLOEXEC
+# ifndef __ASSUME_PIPE2
+  if (__have_pipe2 >= 0)
+# endif
+    {
+      int r = __pipe2 (fildes, O_CLOEXEC);
+# ifndef __ASSUME_PIPE2
+      if (__have_pipe2 == 0)
+	__have_pipe2 = r != -1 || errno != ENOSYS ? 1 : -1;
+
+      if (__have_pipe2 > 0)
+# endif
+	if (r < 0)
+	  /* Bad */
+	  return WRDE_NOSPACE;
+    }
+#endif
+#ifndef __ASSUME_PIPE2
+# ifdef O_CLOEXEC
+  if (__have_pipe2 < 0)
+# endif
+    if (__pipe (fildes) < 0)
+      /* Bad */
+      return WRDE_NOSPACE;
+#endif
 
  again:
-  pid = exec_comm_child (comm, fildes, noexec ? false : flags & WRDE_SHOWERR,
-			 noexec);
-  if (pid < 0)
+  /* It is not safe to use vfork in lieu of fork because the child
+     calls unsetenv to remove "IFS" from the environment.  If we were
+     to use vfork, that would modify the parent's environment.  */
+  if ((pid = __fork ()) < 0)
     {
+      /* Bad */
       __close (fildes[0]);
       __close (fildes[1]);
       return WRDE_NOSPACE;
     }
 
+  if (pid == 0)
+    exec_comm_child (comm, fildes, noexec ? 0 : flags & WRDE_SHOWERR, noexec);
+
+  /* Parent */
+
   /* If we are just testing the syntax, only wait.  */
   if (noexec)
     return (TEMP_FAILURE_RETRY (__waitpid (pid, &status, 0)) == pid
@@ -922,12 +958,7 @@ exec_comm (char *comm, char **word, size_t *word_length, size_t *max_length,
 	  if ((buflen = TEMP_FAILURE_RETRY (__read (fildes[0], buffer,
 						    bufsize))) < 1)
 	    {
-	      /* If read returned 0 then the process has closed its
-		 stdout.  Don't use WNOHANG in that case to avoid busy
-		 looping until the process eventually exits.  */
-	      if (TEMP_FAILURE_RETRY (__waitpid (pid, &status,
-						 buflen == 0 ? 0 : WNOHANG))
-		  == 0)
+	      if (TEMP_FAILURE_RETRY (__waitpid (pid, &status, WNOHANG)) == 0)
 		continue;
 	      if ((buflen = TEMP_FAILURE_RETRY (__read (fildes[0], buffer,
 							bufsize))) < 1)
@@ -957,12 +988,7 @@ exec_comm (char *comm, char **word, size_t *word_length, size_t *max_length,
 	  if ((buflen = TEMP_FAILURE_RETRY (__read (fildes[0], buffer,
 						    bufsize))) < 1)
 	    {
-	      /* If read returned 0 then the process has closed its
-		 stdout.  Don't use WNOHANG in that case to avoid busy
-		 looping until the process eventually exits.  */
-	      if (TEMP_FAILURE_RETRY (__waitpid (pid, &status,
-						 buflen == 0 ? 0 : WNOHANG))
-		  == 0)
+	      if (TEMP_FAILURE_RETRY (__waitpid (pid, &status, WNOHANG)) == 0)
 		continue;
 	      if ((buflen = TEMP_FAILURE_RETRY (__read (fildes[0], buffer,
 							bufsize))) < 1)
@@ -1039,7 +1065,7 @@ exec_comm (char *comm, char **word, size_t *word_length, size_t *max_length,
 		  if (copying == 3)
 		    {
 		      /* Nothing but (IFS) newlines since the last field,
-			 so delimit it here before starting new word */
+		         so delimit it here before starting new word */
 		      if (w_addword (pwordexp, *word) == WRDE_NOSPACE)
 			goto no_space;
 
@@ -1065,8 +1091,8 @@ exec_comm (char *comm, char **word, size_t *word_length, size_t *max_length,
   /* Chop off trailing newlines (required by POSIX.2)  */
   /* Ensure we don't go back further than the beginning of the
      substitution (i.e. remove maxnewlines bytes at most) */
-  while (maxnewlines-- != 0
-	 && *word_length > 0 && (*word)[*word_length - 1] == '\n')
+  while (maxnewlines-- != 0 &&
+         *word_length > 0 && (*word)[*word_length - 1] == '\n')
     {
       (*word)[--*word_length] = '\0';
 
@@ -1086,7 +1112,7 @@ exec_comm (char *comm, char **word, size_t *word_length, size_t *max_length,
   /* Check for syntax error (re-execute but with "-n" flag) */
   if (buflen < 1 && status != 0)
     {
-      noexec = true;
+      noexec = 1;
       goto again;
     }
 
@@ -1100,6 +1126,7 @@ no_space:
 }
 
 static int
+internal_function
 parse_comm (char **word, size_t *word_length, size_t *max_length,
 	    const char *words, size_t *offset, int flags, wordexp_t *pwordexp,
 	    const char *ifs, const char *ifs_white)
@@ -1138,9 +1165,25 @@ parse_comm (char **word, size_t *word_length, size_t *max_length,
 	      /* Go -- give script to the shell */
 	      if (comm)
 		{
-		  /* posix_spawn already handles thread cancellation.  */
+#ifdef __libc_ptf_call
+		  /* We do not want the exec_comm call to be cut short
+		     by a thread cancellation since cleanup is very
+		     ugly.  Therefore disable cancellation for
+		     now.  */
+		  // XXX Ideally we do want the thread being cancelable.
+		  // XXX If demand is there we'll change it.
+		  int state = PTHREAD_CANCEL_ENABLE;
+		  __libc_ptf_call (pthread_setcancelstate,
+				   (PTHREAD_CANCEL_DISABLE, &state), 0);
+#endif
+
 		  error = exec_comm (comm, word, word_length, max_length,
 				     flags, pwordexp, ifs, ifs_white);
+
+#ifdef __libc_ptf_call
+		  __libc_ptf_call (pthread_setcancelstate, (state, NULL), 0);
+#endif
+
 		  free (comm);
 		}
 
@@ -1166,10 +1209,8 @@ parse_comm (char **word, size_t *word_length, size_t *max_length,
   return WRDE_SYNTAX;
 }
 
-#define CHAR_IN_SET(ch, char_set) \
-  (memchr (char_set "", ch, sizeof (char_set) - 1) != NULL)
-
 static int
+internal_function
 parse_param (char **word, size_t *word_length, size_t *max_length,
 	     const char *words, size_t *offset, int flags, wordexp_t *pwordexp,
 	     const char *ifs, const char *ifs_white, int quoted)
@@ -1249,7 +1290,7 @@ parse_param (char **word, size_t *word_length, size_t *max_length,
 	}
       while (isdigit(words[++*offset]));
     }
-  else if (CHAR_IN_SET (words[*offset], "*@$"))
+  else if (strchr ("*@$", words[*offset]) != NULL)
     {
       /* Special parameter. */
       special = 1;
@@ -1293,7 +1334,7 @@ parse_param (char **word, size_t *word_length, size_t *max_length,
 	  break;
 
 	case ':':
-	  if (!CHAR_IN_SET (words[1 + *offset], "-=?+"))
+	  if (strchr ("-=?+", words[1 + *offset]) == NULL)
 	    goto syntax;
 
 	  colon_seen = 1;
@@ -1398,7 +1439,7 @@ envsubst:
   /* Is it a numeric parameter? */
   else if (isdigit (env[0]))
     {
-      unsigned long n = strtoul (env, NULL, 10);
+      int n = atoi (env);
 
       if (n >= __libc_argc)
 	/* Substitute NULL. */
@@ -1539,7 +1580,7 @@ envsubst:
       if (expand_pattern)
 	{
 	  /* We need to perform tilde expansion, parameter expansion,
-	     command substitution, and arithmetic expansion.  We also
+             command substitution, and arithmetic expansion.  We also
 	     have to be a bit careful with wildcard characters, as
 	     pattern might be given to fnmatch soon.  To do this, we
 	     convert quotes to escapes. */
@@ -1790,7 +1831,7 @@ envsubst:
 	    {
 	      const char *str = pattern;
 
-	      if (!str || str[0] == '\0')
+	      if (str[0] == '\0')
 		str = _("parameter null or not set");
 
 	      __fxprintf (NULL, "%s: %s\n", env, str);
@@ -1862,7 +1903,7 @@ envsubst:
 	  if (pattern && !value)
 	    goto no_space;
 
-	  __setenv (env, value ?: "", 1);
+	  __setenv (env, value, 1);
 	  break;
 
 	default:
@@ -1883,7 +1924,10 @@ envsubst:
 			_itoa_word (value ? strlen (value) : 0,
 				    &param_length[20], 10, 0));
       if (free_value)
-	free (value);
+	{
+	  assert (value != NULL);
+	  free (value);
+	}
 
       return *word ? 0 : WRDE_NOSPACE;
     }
@@ -1941,7 +1985,7 @@ envsubst:
 	  field_end = field_begin + strcspn (field_begin, ifs);
 
 	  /* Set up pointer to the character after end of field and
-	     skip whitespace IFS after it. */
+             skip whitespace IFS after it. */
 	  next_field = field_end + strspn (field_end, ifs_white);
 
 	  /* Skip at most one non-whitespace IFS character after the field */
@@ -1992,9 +2036,8 @@ do_error:
   return error;
 }
 
-#undef CHAR_IN_SET
-
 static int
+internal_function
 parse_dollars (char **word, size_t *word_length, size_t *max_length,
 	       const char *words, size_t *offset, int flags,
 	       wordexp_t *pwordexp, const char *ifs, const char *ifs_white,
@@ -2034,6 +2077,9 @@ parse_dollars (char **word, size_t *word_length, size_t *max_length,
 	    }
 	}
 
+      if (flags & WRDE_NOCMD)
+	return WRDE_CMDSUB;
+
       (*offset) += 2;
       return parse_comm (word, word_length, max_length, words, offset, flags,
 			 quoted? NULL : pwordexp, ifs, ifs_white);
@@ -2053,6 +2099,7 @@ parse_dollars (char **word, size_t *word_length, size_t *max_length,
 }
 
 static int
+internal_function
 parse_backtick (char **word, size_t *word_length, size_t *max_length,
 		const char *words, size_t *offset, int flags,
 		wordexp_t *pwordexp, const char *ifs, const char *ifs_white)
@@ -2090,6 +2137,7 @@ parse_backtick (char **word, size_t *word_length, size_t *max_length,
 	      break;
 	    }
 
+	  ++(*offset);
 	  error = parse_backslash (&comm, &comm_length, &comm_maxlen, words,
 				   offset);
 
@@ -2103,7 +2151,6 @@ parse_backtick (char **word, size_t *word_length, size_t *max_length,
 
 	case '\'':
 	  squoting = 1 - squoting;
-	  [[fallthrough]];
 	default:
 	  comm = w_addchar (comm, &comm_length, &comm_maxlen, words[*offset]);
 	  if (comm == NULL)
@@ -2117,6 +2164,7 @@ parse_backtick (char **word, size_t *word_length, size_t *max_length,
 }
 
 static int
+internal_function
 parse_dquote (char **word, size_t *word_length, size_t *max_length,
 	      const char *words, size_t *offset, int flags,
 	      wordexp_t *pwordexp, const char * ifs, const char * ifs_white)
@@ -2143,6 +2191,9 @@ parse_dquote (char **word, size_t *word_length, size_t *max_length,
 	  break;
 
 	case '`':
+	  if (flags & WRDE_NOCMD)
+	    return WRDE_CMDSUB;
+
 	  ++(*offset);
 	  error = parse_backtick (word, word_length, max_length, words,
 				  offset, flags, NULL, NULL, NULL);
@@ -2216,9 +2267,7 @@ wordexp (const char *words, wordexp_t *pwordexp, int flags)
     {
       /* Minimal implementation of WRDE_REUSE for now */
       wordfree (pwordexp);
-      old_word.we_wordc = 0;
       old_word.we_wordv = NULL;
-      pwordexp->we_wordc = 0;
     }
 
   if ((flags & WRDE_APPEND) == 0)
@@ -2303,6 +2352,12 @@ wordexp (const char *words, wordexp_t *pwordexp, int flags)
 	break;
 
       case '`':
+	if (flags & WRDE_NOCMD)
+	  {
+	    error = WRDE_CMDSUB;
+	    goto do_error;
+	  }
+
 	++words_offset;
 	error = parse_backtick (&word, &word_length, &max_length, words,
 				&words_offset, flags, pwordexp, ifs,
