@@ -1,5 +1,5 @@
 /* Byte-wise substring search, using the Two-Way algorithm.
-   Copyright (C) 2008-2019 Free Software Foundation, Inc.
+   Copyright (C) 2008, 2010 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
    Written by Eric Blake <ebb9@byu.net>, 2008.
 
@@ -14,12 +14,13 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the GNU C Library; if not, see
-   <http://www.gnu.org/licenses/>.  */
+   License along with the GNU C Library; if not, write to the Free
+   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   02111-1307 USA.  */
 
 /* Before including this file, you need to include <string.h> (and
    <config.h> before that, if not part of libc), and define:
-     RETURN_TYPE             A macro that expands to the return type.
+     RESULT_TYPE             A macro that expands to the return type.
      AVAILABLE(h, h_l, j, n_l)
 			     A macro that returns nonzero if there are
 			     at least N_L bytes left starting at H[J].
@@ -37,17 +38,12 @@
 			     The argument is an 'unsigned char'; the result
 			     must be an 'unsigned char' as well.
 
-  Other macros you may optionally define:
-     RET0_IF_0(a)            Documented below at default definition.
-     CHECK_EOL               Same.
-
-  This file undefines the macros listed above, and defines
+  This file undefines the macros documented above, and defines
   LONG_NEEDLE_THRESHOLD.
 */
 
 #include <limits.h>
 #include <stdint.h>
-#include <sys/param.h>                  /* Defines MAX.  */
 
 /* We use the Two-Way string matching algorithm, which guarantees
    linear complexity with constant space.  Additionally, for long
@@ -72,6 +68,10 @@
 # define LONG_NEEDLE_THRESHOLD SIZE_MAX
 #endif
 
+#ifndef MAX
+# define MAX(a, b) ((a < b) ? (b) : (a))
+#endif
+
 #ifndef CANON_ELEMENT
 # define CANON_ELEMENT(c) c
 #endif
@@ -79,17 +79,18 @@
 # define CMP_FUNC memcmp
 #endif
 
-/* Check for end-of-line in strstr and strcasestr routines.
-   We piggy-back matching procedure for detecting EOL where possible,
-   and use AVAILABLE macro otherwise.  */
-#ifndef CHECK_EOL
-# define CHECK_EOL (0)
+#ifndef AVAILABLE1
+# define AVAILABLE1(h, h_l, j, n_l) AVAILABLE (h, h_l, j, n_l)
+#endif
+#ifndef AVAILABLE2
+# define AVAILABLE2(h, h_l, j, n_l) (1)
 #endif
-
-/* Return NULL if argument is '\0'.  */
 #ifndef RET0_IF_0
 # define RET0_IF_0(a) /* nothing */
 #endif
+#ifndef AVAILABLE_USES_J
+# define AVAILABLE_USES_J (1)
+#endif
 
 /* Perform a critical factorization of NEEDLE, of length NEEDLE_LEN.
    Return the index of the first byte in the right half, and set
@@ -221,7 +222,7 @@ critical_factorization (const unsigned char *needle, size_t needle_len,
    most 2 * HAYSTACK_LEN - NEEDLE_LEN comparisons occur in searching.
    If AVAILABLE modifies HAYSTACK_LEN (as in strstr), then at most 3 *
    HAYSTACK_LEN - NEEDLE_LEN comparisons occur in searching.  */
-static inline RETURN_TYPE
+static RETURN_TYPE
 two_way_short_needle (const unsigned char *haystack, size_t haystack_len,
 		      const unsigned char *needle, size_t needle_len)
 {
@@ -281,7 +282,7 @@ two_way_short_needle (const unsigned char *haystack, size_t haystack_len,
     }
   else
     {
-      const unsigned char *phaystack;
+      const unsigned char *phaystack = &haystack[suffix];
       /* The comparison always starts from needle[suffix], so cache it
 	 and use an optimized first-character loop.  */
       unsigned char needle_suffix = CANON_ELEMENT (needle[suffix]);
@@ -290,59 +291,42 @@ two_way_short_needle (const unsigned char *haystack, size_t haystack_len,
 	 required, and any mismatch results in a maximal shift.  */
       period = MAX (suffix, needle_len - suffix) + 1;
       j = 0;
-      while (AVAILABLE (haystack, haystack_len, j, needle_len))
+      while (AVAILABLE1 (haystack, haystack_len, j, needle_len))
 	{
-	  unsigned char haystack_char;
 	  const unsigned char *pneedle;
+	  unsigned char a;
 
-	  phaystack = &haystack[suffix + j];
-
-#ifdef FASTSEARCH
-	  if (*phaystack++ != needle_suffix)
+	  /* TODO: The first-character loop can be sped up by adapting
+	     longword-at-a-time implementation of memchr/strchr.  */
+	  if (needle_suffix
+	      != (a = CANON_ELEMENT (*phaystack++)))
 	    {
-	      phaystack = FASTSEARCH (phaystack, needle_suffix,
-				      haystack_len - needle_len - j);
-	      if (phaystack == NULL)
-		goto ret0;
-	      j = phaystack - &haystack[suffix];
-	      phaystack++;
-	    }
-#else
-	  while (needle_suffix
-	      != (haystack_char = CANON_ELEMENT (*phaystack++)))
-	    {
-	      RET0_IF_0 (haystack_char);
-# if !CHECK_EOL
+#if AVAILABLE_USES_J
 	      ++j;
-	      if (!AVAILABLE (haystack, haystack_len, j, needle_len))
-		goto ret0;
-# endif
+#endif
+	      RET0_IF_0 (a);
+	      continue;
 	    }
 
-# if CHECK_EOL
+#if !AVAILABLE_USES_J
 	  /* Calculate J if it wasn't kept up-to-date in the first-character
 	     loop.  */
 	  j = phaystack - &haystack[suffix] - 1;
-# endif
 #endif
+
 	  /* Scan for matches in right half.  */
 	  i = suffix + 1;
 	  pneedle = &needle[i];
 	  while (i < needle_len)
 	    {
 	      if (CANON_ELEMENT (*pneedle++)
-		  != (haystack_char = CANON_ELEMENT (*phaystack++)))
+		  != (a = CANON_ELEMENT (*phaystack++)))
 		{
-		  RET0_IF_0 (haystack_char);
+		  RET0_IF_0 (a);
 		  break;
 		}
 	      ++i;
 	    }
-#if CHECK_EOL
-	  /* Update minimal length of haystack.  */
-	  if (phaystack > haystack + haystack_len)
-	    haystack_len = phaystack - haystack;
-#endif
 	  if (needle_len <= i)
 	    {
 	      /* Scan for matches in left half.  */
@@ -352,9 +336,9 @@ two_way_short_needle (const unsigned char *haystack, size_t haystack_len,
 	      while (i != SIZE_MAX)
 		{
 		  if (CANON_ELEMENT (*pneedle--)
-		      != (haystack_char = CANON_ELEMENT (*phaystack--)))
+		      != (a = CANON_ELEMENT (*phaystack--)))
 		    {
-		      RET0_IF_0 (haystack_char);
+		      RET0_IF_0 (a);
 		      break;
 		    }
 		  --i;
@@ -365,6 +349,11 @@ two_way_short_needle (const unsigned char *haystack, size_t haystack_len,
 	    }
 	  else
 	    j += i - suffix + 1;
+
+	  if (!AVAILABLE2 (haystack, haystack_len, j, needle_len))
+	    break;
+
+	  phaystack = &haystack[suffix + j];
 	}
     }
  ret0: __attribute__ ((unused))
@@ -382,11 +371,8 @@ two_way_short_needle (const unsigned char *haystack, size_t haystack_len,
    and sublinear performance O(HAYSTACK_LEN / NEEDLE_LEN) is possible.
    If AVAILABLE modifies HAYSTACK_LEN (as in strstr), then at most 3 *
    HAYSTACK_LEN - NEEDLE_LEN comparisons occur in searching, and
-   sublinear performance is not possible.
-
-   Since this function is large and complex, block inlining to avoid
-   slowing down the common case of small needles.  */
-__attribute__((noinline)) static RETURN_TYPE
+   sublinear performance is not possible.  */
+static RETURN_TYPE
 two_way_long_needle (const unsigned char *haystack, size_t haystack_len,
 		     const unsigned char *needle, size_t needle_len)
 {
@@ -517,12 +503,15 @@ two_way_long_needle (const unsigned char *haystack, size_t haystack_len,
 	    j += i - suffix + 1;
 	}
     }
+ ret0: __attribute__ ((unused))
   return NULL;
 }
 
 #undef AVAILABLE
+#undef AVAILABLE1
+#undef AVAILABLE2
+#undef AVAILABLE_USES_J
 #undef CANON_ELEMENT
 #undef CMP_FUNC
 #undef RET0_IF_0
 #undef RETURN_TYPE
-#undef CHECK_EOL
