@@ -1,4 +1,4 @@
-/* Copyright (C) 2002-2018 Free Software Foundation, Inc.
+/* Copyright (C) 2002, 2003, 2005 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
 
    The GNU C Library is free software; you can redistribute it and/or
@@ -12,8 +12,9 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the GNU C Library; if not, see
-   <http://www.gnu.org/licenses/>.  */
+   License along with the GNU C Library; if not, write to the Free
+   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   02111-1307 USA.  */
 
 #include <sched.h>
 #include <signal.h>
@@ -21,7 +22,8 @@
 #include <sysdep.h>
 #include <unistd.h>
 #include <sys/wait.h>
-#include <libc-lock.h>
+#include <bits/libc-lock.h>
+#include <kernel-features.h>
 
 /* We have to and actually can handle cancelable system().  The big
    problem: we have to kill the child process if necessary.  To do
@@ -30,20 +32,45 @@
    the PID when needed.  It is not necessary for the parent thread to
    return.  It might still be in the kernel when the cancellation
    request comes.  Therefore we have to use the clone() calls ability
-   to have the kernel write the PID into the user-level variable.  */
+   to have the kernel write the PID into the user-level variable.  
+
+   We must use cleanup_id, instead of the existing pid variable, when
+   using vfork, so that we do not overwrite the parent's pid variable.
+   Otherwise, if the child execs, the parent is cancelled, but vfork
+   has not yet returned in the parent, the pid variable will have the
+   value zero, not the child's PID.  */
 #ifndef FORK
-# define FORK() \
-  INLINE_SYSCALL (clone, 3, CLONE_PARENT_SETTID | SIGCHLD, 0, &pid)
+/* Use vfork so that if running without overcommit we do not make a
+   copy of the parent's memory.  */
+# define FORK_IS_VFORK
+# if defined __ASSUME_CLONE_THREAD_FLAGS	
+#  define USE_CLEANUP_PID
+#  define FORK()						\
+  INLINE_SYSCALL (clone, 3,					\
+                  CLONE_PARENT_SETTID | CLONE_VFORK | SIGCHLD,	\
+                  0, &cleanup_pid)
+# else
+#  define FORK() __vfork()
+# endif
 #endif
 
 #ifdef _LIBC_REENTRANT
 static void cancel_handler (void *arg);
 
-# define CLEANUP_HANDLER \
-  __libc_cleanup_region_start (1, cancel_handler, &pid)
+#ifdef USE_CLEANUP_PID
+# define CLEANUP_HANDLER						\
+  {									\
+    pid_t cleanup_pid;							\
+    __libc_cleanup_region_start (1, cancel_handler, &cleanup_pid)
+#else
+# define CLEANUP_HANDLER					\
+  {								\
+    __libc_cleanup_region_start (1, cancel_handler, &pid)
+#endif
 
-# define CLEANUP_RESET \
-  __libc_cleanup_region_end (0)
+# define CLEANUP_RESET				\
+    __libc_cleanup_region_end (0);		\
+  }
 #endif
 
 
